/* Musical Blob
 * University of Washington Computer Science and Engineering
 * Sound Capstone (CSE 481i)
 * Project Members: Nathan Armstrong, Daniel Chesney, Andrew Foster
 *
 * BlobSurfaceDrawer
 * For documentation, see BlobIsoFunc.hpp
 */

//Evaluate the isolevel at the given evaluation point, with the given source
//point generating the field.  The isolevel is a function of distance from the
//source point, and was developed by Ken Perlin, as seen on this page:
//  http://www.geisswerks.com/ryan/BLOBS/blobs.html

#include <omp.h>
#ifdef __APPLE__
#include <GLUT/glut.h>
#include <stdlib.h>
#else
#include <GL/glut.h>
//#if defined(linux) || defined(__linux)
//#include <stdlib.h>
//#endif
#endif
#include <stdio.h>
#include <algorithm>
#include "BlobIsoFunc.hpp"

//Isosurface level (distance from a lone particle at which the isosurface is
//expected to be seen)
#define ISOLEV 100

//Extra margin to add to the isolevel when partitioning particles to ensure
//complete rendering.  Unnecessary now since isosurface is always guaranteed to
//be within the sphere defined by a lone particle's isosurface.
#define MARGIN 0

//Transparency of the rendered surfaces
#define ALPHALEV 0.6

using std::min;
using std::max;

//A set of colors used for surface lighting
static const float afAmbientWhite [] = {0.25, 0.25, 0.25, ALPHALEV};
static const float afAmbientRed   [] = {0.25, 0.00, 0.00, ALPHALEV};
static const float afAmbientGreen [] = {0.00, 0.25, 0.00, ALPHALEV};
static const float afAmbientBlue  [] = {0.00, 0.00, 0.25, ALPHALEV};
static const float afDiffuseWhite [] = {0.75, 0.75, 0.75, ALPHALEV};
static const float afDiffuseRed   [] = {0.75, 0.00, 0.00, ALPHALEV};
static const float afDiffuseGreen [] = {0.00, 0.75, 0.00, ALPHALEV};
static const float afDiffuseBlue  [] = {0.00, 0.00, 0.75, ALPHALEV};
static const float afSpecularWhite[] = {1.00, 1.00, 1.00, ALPHALEV};
static const float afSpecularRed  [] = {1.00, 0.25, 0.25, ALPHALEV};
static const float afSpecularGreen[] = {0.25, 1.00, 0.25, ALPHALEV};
static const float afSpecularBlue [] = {0.25, 0.25, 1.00, ALPHALEV};

BlobSurfaceDrawer::BlobSurfaceDrawer(int xCells, int yCells, int zCells)
									 : CIsoSurface(), xCells(xCells),
									   yCells(yCells), zCells(zCells){
	isolevBuffer = new double[(xCells + 1) * (yCells + 1) * (zCells + 1)];
}

BlobSurfaceDrawer::~BlobSurfaceDrawer(){
	delete[] isolevBuffer;
}

//invIsoLev here is the inverse of the desired distance for the isosurface,
//used to scale the results properly.  Returns true iff a result is within
//potentially reasonable bounds.  A return value of 0 means the point lies on
//or outside the isosurface.
double BlobSurfaceDrawer::evalPointIsoLevel(double invIsoLev,
	                                        const TripleD& source,
											const TripleD& evalPoint){
	TripleD diff = evalPoint - source;
	double rad = 1 - diff.magnitude() * invIsoLev;
	if(rad <= 0) return	.01;
	return -rad * rad * rad * (rad * (rad * 6 - 15) + 10);
}

//Obtains the sum isolevel at the given evaluation point, given all the points
//in the given particle system as the field source.
double BlobSurfaceDrawer::sumIsoLevels(
		const vector<Particle*>& parts, TripleD& evalPoint,
		double invIsoLev){
	double isoLevelSum = 0.0;
	vector<Particle*>::const_iterator i;
	for(i = parts.begin(); i != parts.end(); i++){
		double lev =
			evalPointIsoLevel(invIsoLev, (*i)->getCurrPos(), evalPoint);
		isoLevelSum += lev;
	}
	return isoLevelSum;
}

void BlobSurfaceDrawer::isolevArrGen(
		const vector<Particle*>& parts, TripleD minBounds,
	    double xWidth, double yWidth, double zWidth, double isoLev){
	int zSliceCells = (xCells+1) * (yCells+1);
	double invIsoLev = 1 / isoLev;
#pragma omp parallel for //collapse(3)
	for(int z = 0; z <= zCells; z++){
		for(int y = 0; y <= yCells; y++){
			for(int x = 0; x <= xCells; x++){
				int arrIndex = x + (y * (xCells+1)) + (z * zSliceCells);
				TripleD evalP(x * xWidth, y * yWidth, z * zWidth);
				evalP += minBounds;
				double arrVal = sumIsoLevels(parts, evalP, invIsoLev);
				isolevBuffer[arrIndex] = arrVal;
			}
		}
	}

	for(int z = 0; z <= zCells; z++){
		for(int y = 0; y <= yCells; y++){
			for(int x = 0; x <= xCells; x++){
				int arrIndex = x + (y * (xCells+1)) + (z * zSliceCells);
				TripleD evalP(x * xWidth, y * yWidth, z * zWidth);
				evalP += minBounds;
				double arrVal = sumIsoLevels(parts, evalP, invIsoLev);
				isolevBuffer[arrIndex] = arrVal;
			}
		}
	}
}

bool BlobSurfaceDrawer::generateBlobSurface(
		const vector<Particle*>& parts,
		TripleD minBounds, TripleD maxBounds){
	TripleD boundsDim = maxBounds - minBounds;
	double xCellWidth = boundsDim.getElemA() / xCells;
	double yCellWidth = boundsDim.getElemB() / yCells;
	double zCellWidth = boundsDim.getElemC() / zCells;
	isolevArrGen(parts, minBounds, xCellWidth, yCellWidth, zCellWidth, ISOLEV);

	GenerateSurface(isolevBuffer, 0, xCells, yCells, zCells,
		            (float)xCellWidth, (float)yCellWidth, (float)zCellWidth);
	return IsSurfaceValid();
}

void BlobSurfaceDrawer::draw(const ParticleSystem* ps){
		glDisable(GL_LIGHTING);

	    glColor4f(0.0, 0.0, 0.0, 1.0);
	    std::vector<Particle*> parts = ps->getParticles();
		std::vector<Particle*>::const_iterator i;
	    glBegin(GL_POINTS);
	 	    for(i = parts.begin(); i != parts.end(); i++){
			    TripleD pos = (*i)->getCurrPos();
			    //glVertex3d(pos.getElemA(), pos.getElemB(), pos.getElemC());
			    glVertex3d(pos.getElemA(), pos.getElemB(), 0);
			    glVertex3d(0, pos.getElemB(), pos.getElemC());
			    glVertex3d(pos.getElemA(), 0, pos.getElemC());
		    }
	    glEnd();

		/*for(i = parts.begin(); i != parts.end(); i++){
			TripleD pos = (*i)->getCurrPos();
			glPushMatrix();
				glTranslated(pos.getElemA(), pos.getElemB(), pos.getElemC());
				glutWireSphere(20, 20, 20);
			glPopMatrix();
		}*/

		glEnable(GL_LIGHTING);
	   	glMaterialfv(GL_BACK,  GL_AMBIENT,   afAmbientGreen);
		glMaterialfv(GL_BACK,  GL_DIFFUSE,   afDiffuseGreen);
		glMaterialfv(GL_FRONT, GL_AMBIENT,   afAmbientBlue);
		glMaterialfv(GL_FRONT, GL_DIFFUSE,   afDiffuseBlue);
		glMaterialfv(GL_FRONT, GL_SPECULAR,  afSpecularWhite);
		glMaterialf( GL_FRONT, GL_SHININESS, 25.0);


		vector<ParticleBoundSet*> psets =
			partitionParticles(ps, MARGIN + ISOLEV);
		vector<ParticleBoundSet*>::const_iterator j;
		for(j = psets.begin(); j != psets.end(); j++){
			ParticleBoundSet* pbs = *j;
			TripleD minBound = pbs->getMinBound();
			if(generateBlobSurface(pbs->getParts(), minBound,
			                       pbs->getMaxBound())){
				glBegin(GL_TRIANGLES);
				for(unsigned int i = 0; i < (m_nTriangles * 3); i++){
					POINT3D& v = m_ppt3dVertices[m_piTriangleIndices[i]];
					VECTOR3D& n = m_pvec3dNormals[m_piTriangleIndices[i]];
					glNormal3f(n[0], n[1], n[2]);
					glVertex3f(v[0] + (float)minBound.getElemA(),
					           v[1] + (float)minBound.getElemB(),
							   v[2] + (float)minBound.getElemC());
				}
				glEnd();
			}

		/* DISPLAY TEST CUBES */

			/*const TripleD maxBound = (*j)->getMaxBound();
			TripleD dim = maxBound - minBound;

			if(   dim.getElemA() == 0
			   || dim.getElemB() == 0
			   || dim.getElemC() == 0) continue;

			glPushMatrix();
				glTranslated(minBound.getElemA(),
					         minBound.getElemB(),
							 minBound.getElemC());
				glScaled(dim.getElemA(),
				         dim.getElemB(),
						 dim.getElemC());
				glTranslated(0.5, 0.5, 0.5);
				glutWireCube(1.0);
			glPopMatrix();*/
			delete (*j);
		}

		glDisable(GL_LIGHTING);
	//}
}

vector<ParticleBoundSet*> BlobSurfaceDrawer::
		partitionParticles(const ParticleSystem* ps, double margin){
	vector<Particle*> parts = ps->getParticles();
	ParticleBoundSet** setArr = new ParticleBoundSet*[parts.size()];
	for(unsigned int i = 0; i < parts.size(); i++)
		setArr[i] = new ParticleBoundSet(parts[i], margin);
	int arrSize = parts.size();
	bool changed;
	do{
		changed = false;
		int removed = 0, index = 0;
		for(int i = 0; i < arrSize; i++){
			for(int j = i+1; j < arrSize; j++){
				if(setArr[i]->intersects(setArr[j])){
					setArr[j]->mergeSubset(setArr[i]);
					delete setArr[i];
					setArr[i] = NULL;
					changed = true;
					removed++;
					break;
				}
			}
			if(setArr[i] != NULL){
				setArr[index] = setArr[i];
				index++;
			}
		}
		arrSize -= removed;
	}while(changed);

	vector<ParticleBoundSet*> sets;
	for(int i = 0; i < arrSize; i++){
		sets.push_back(setArr[i]);
	}

	delete[] setArr;
	setArr = NULL;

	return sets;
}
