/* Musical Blob
 * University of Washington Computer Science and Engineering
 * Sound Capstone (CSE 481i)
 * Project Members: Nathan Armstrong, Daniel Chesney, Andrew Foster
 *
 * InputManager
 * For documentation, see InputManager.hpp
 */

#include <map>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <limits>
#include "InputManager.hpp"

using std::map;
using std::pair;

/* LimbGrasp
 * A storage class for holding data about a specific instance of a limb
 * grasping a particle.
 */
class InputManager::LimbGrasp{
public:
	LimbGrasp(unsigned user, LimbType type, Particle* p) :
		id(user), limb(type), part(p) {}

	const unsigned id;
	const LimbType limb;
	Particle* part;
};

InputManager::InputManager(ParticleSystem* ps, unsigned kinectPort,
                           PointMapper mapper) :
		partSys(ps), lt(), kih(kinectPort, &lt), wm(), pm(mapper) {
	wm.connectWiiMotes();
}

void InputManager::listen(){ kih.listen(); }

void InputManager::update(){
	const set<unsigned>& kinectUsers = lt.getUserIDs();
	for(set<unsigned>::const_iterator i = kinectUsers.begin();
	    i != kinectUsers.end(); i++){
		unsigned userID = *i;
		if(lt.limbSetExists(userID)){
			for(unsigned j = 0; j < NUM_LIMBS; j++){
				TripleD limbP, limbV;
				LimbType limb = static_cast<LimbType>(j);
				bool limbDataAvail =
					lt.getLimbPos(userID, limb, limbP) &&
				    lt.getLimbVel(*i, limb, limbV);
				if(limbDataAvail && !isLimbGrasped(userID, limb)){
					applyContactForce(pm.mapPoint(limbP),
					                  pm.scalePoint(limbV));
				}
			}
		}
	}

	wm.update();
	for(unsigned i = 0; i < wm.getNumWiiMotes(); i++){
		if(!(wm.isWiiMoteValid(i) || wm.eventOccurred(i))) continue;
		if(wm.isLimbAttached(i)){
			if(!lt.limbSetExists(wm.getLimbUser(i))){
				wm.detachLimb(i);
				printf("-- wiimote %d detached\n", i);
			} else handleWiiMoteInput(i);
		} else testWiiMoteAttachment(i);
	}

	for(map<unsigned, LimbGrasp*>::iterator i = graspMap.begin();
			i != graspMap.end(); i++){
		LimbGrasp* grasp = (*i).second;
		TripleD limbPos;
		if(lt.getLimbPos(grasp->id, grasp->limb, limbPos))
			grasp->part->setCurrPos(pm.mapPoint(limbPos));
	}
}

void InputManager::drawLimbReps() const {
	static const unsigned NUM_COLORS = 5;
	static const GLubyte userColors[NUM_COLORS][3] = {
		{ 200, 200, 0 },
		{ 128, 0, 180 },
		{ 200, 0, 0 },
		{ 0, 200, 200 },
		{ 0, 200, 0 }
	};

	const set<unsigned>& kinectUsers = lt.getUserIDs();
	unsigned currUser = 0;
	for(set<unsigned>::const_iterator i = kinectUsers.begin();
			i != kinectUsers.end(); i++){
		unsigned userID = *i;
		glColor3ubv(userColors[userID % NUM_COLORS]);
		if(lt.limbSetExists(userID)){
			for(unsigned j = 0; j < NUM_LIMBS; j++){
				TripleD limbPos;
				LimbType limb = static_cast<LimbType>(j);
				if(!lt.getLimbPos(userID, limb, limbPos)) continue;
				limbPos = pm.mapPoint(limbPos);
				glPushMatrix();
					glTranslated(limbPos.getElemA(), limbPos.getElemB(), limbPos.getElemC());
					glutSolidSphere(50.0, 20, 20);
				glPopMatrix();
			}
		}
		currUser++;
	}
}

void InputManager::applyContactForce(TripleD limbP, TripleD limbV){
	vector<Particle*> parts = partSys->getParticles();
	for(vector<Particle*>::const_iterator i = parts.begin();
			i != parts.end(); i++){
		Particle* p = *i;
		TripleD pos = p->getCurrPos();
		TripleD diff = pos - limbP;
		double distMull = (100 - diff.magnitude()) / 10;
		if(distMull > 0)
			p->addForce(diff.normalize().scale(
			              diff * limbV * distMull));
	}
}

void InputManager::applyAttractForce(TripleD sourcePos){
	vector<Particle*> parts = partSys->getParticles();
	for(vector<Particle*>::const_iterator i = parts.begin();
			i != parts.end(); i++){
		Particle* p = *i;
		TripleD pos = p->getCurrPos();
		TripleD diff = sourcePos - pos;
		if(diff.magnitude() < 500.0)
			p->addForce(diff.normalize().scale(100));
	}
}

bool InputManager::isLimbGrasped(unsigned userID, LimbType limb) const{
	for(map<unsigned, LimbGrasp*>::const_iterator i = graspMap.begin();
			i != graspMap.end(); i++){
		LimbGrasp* grasp = (*i).second;
		if(grasp->id == userID && grasp->limb == limb) return true;
	}
	return false;
}

bool InputManager::isParticleGrasped(Particle* p) const{
	for(map<unsigned, LimbGrasp*>::const_iterator i = graspMap.begin();
			i != graspMap.end(); i++){
		LimbGrasp* grasp = (*i).second;
		if(grasp->part == p) return true;
	}
	return false;
}

void InputManager::handleWiiMoteInput(unsigned mote){
	unsigned userID = wm.getLimbUser(mote);
	LimbType limb = wm.getLimb(mote);
	TripleD limbPos;
	if(!lt.getLimbPos(userID, limb, limbPos)) return;
	limbPos = pm.mapPoint(limbPos);

	if(wm.isButtonHeld(mote, WIIMOTE_BUTTON_B)
	   && graspMap.find(mote) == graspMap.end()){
		//User is attempting to grasp the blob.
		double minDist = 150.0;
		Particle* minPart = NULL;
		vector<Particle*> parts = partSys->getParticles();
		for(vector<Particle*>::const_iterator i = parts.begin();
				i != parts.end(); i++){
			TripleD partPos = (*i)->getCurrPos();
			double dist = (partPos - limbPos).magnitude();
			if(dist < minDist && !isParticleGrasped(*i)){
				minDist = dist;
				minPart = *i;
			}
		}
		if(minPart != NULL){
			LimbGrasp* grasp = new LimbGrasp(userID, limb, minPart);
			graspMap.insert(pair<unsigned, LimbGrasp*>(mote, grasp));
			minPart->setPhysicsActive(false);
			printf("-- wiimote %d initiating grasp\n", mote);
		}
	}else if(!wm.isButtonHeld(mote, WIIMOTE_BUTTON_B)){
		//User is attempting to release the blob.
		map<unsigned, LimbGrasp*>::iterator i = graspMap.find(mote);
		if(i != graspMap.end()){
			LimbGrasp* grasp = i->second;
			graspMap.erase(mote);
			grasp->part->setPhysicsActive(true);
			grasp->part->setPrevPos(grasp->part->getCurrPos());
			delete grasp;
			printf("-- wiimote %d releasing grasp\n", mote);
		}
	}
	if(wm.isButtonHeld(mote, WIIMOTE_BUTTON_A))
		applyAttractForce(limbPos);
}

void InputManager::testWiiMoteAttachment(unsigned mote){
	if(!wm.isButtonPressed(mote, WIIMOTE_BUTTON_HOME)) return;

	set<unsigned> ids = lt.getUserIDs();
	double minY = std::numeric_limits<double>::max();
	int minID = -1;
	LimbType minLimb = NUM_LIMBS;

	//Find the limb that is highest (minimum Y value)
	for(set<unsigned>::iterator i = ids.begin(); i != ids.end(); i++){
		for(unsigned j = 0; j < NUM_LIMBS; j++){
			TripleD limbPos;
			LimbType limb = static_cast<LimbType>(j);
			if(!lt.getLimbPos(*i, limb, limbPos)) continue;
			if(limbPos.getElemB() < minY){
				minY = limbPos.getElemB();
				minID = *i;
				minLimb = limb;
			}
		}
	}
	if(minID != -1){
		wm.attachLimb(mote, minID, minLimb);
		printf("-- wiimote %d attached to user %d limb %d\n", mote, minID, minLimb);
	}
}
