/* Musical Blob
 * University of Washington Computer Science and Engineering
 * Sound Capstone (CSE 481i)
 * Project Members: Nathan Armstrong, Daniel Chesney, Andrew Foster
 *
 * WiiManager
 * For documentation, see WiiManager.hpp
 */

#include "WiiManager.hpp"

//for usleep
#include <unistd.h>
#include <algorithm>

using std::min;

//LimbAttachment
//Support class for holding data concerning WiiMote attachment to a limb
class WiiManager::LimbAttachment{
public:
	//Creates a new LimbAttachment, with no limb attached.
	LimbAttachment() : limbAttached(false), limb(NUM_LIMBS) {}

	//Returns whether this represents an attached limb
	bool isAttached() const { return limbAttached; }

	//Returns the user ID of this limb attachment
	//Only valid if isAttached returns true
	unsigned getAttachedUser() const { return userID; }

	//Returns the limb of this limb attachment
	//Only valid if isAttached returns true
	LimbType getAttachedLimb() const { return limb; }

	//Sets this limb attachment to the given user ID and limb
	void setLimbAttachment(unsigned id, LimbType type){
		limbAttached = true;
		userID = id;
		limb = type;
	}

	//Returns this LimbAttachment to an unattached state
	void removeLimbAttachment(){
		limbAttached = false;
	}

private:
	//User limb attachment
	bool limbAttached;
	unsigned userID;
	LimbType limb;
};

WiiManager::WiiManager(){
	motes = wiiuse_init(MAX_REMOTES);
	for(int i = 0; i < MAX_REMOTES; i++)
		limbs[i] = new LimbAttachment();
}

WiiManager::~WiiManager(){
	wiiuse_cleanup(motes, MAX_REMOTES);
	for(int i = 0; i < MAX_REMOTES; i++)
		delete limbs[i];
}

unsigned WiiManager::getNumWiiMotes() const {
	return numRemotes;
}

unsigned WiiManager::getLimbUser(unsigned mote) const {
	return limbs[mote]->getAttachedUser();
}

LimbType WiiManager::getLimb(unsigned mote) const {
	return limbs[mote]->getAttachedLimb();
}

bool WiiManager::isLimbAttached(unsigned mote) const {
	return mote < numRemotes && moteValid[mote] &&
	       limbs[mote]->isAttached();
}

bool WiiManager::isWiiMoteValid(unsigned mote) const {
	return moteValid[mote];
}

bool WiiManager::eventOccurred(unsigned mote) const {
	return isWiiMoteValid(mote) && motes[mote]->event == WIIUSE_EVENT;
}

bool WiiManager::isButtonHeld(unsigned mote, unsigned button) const {
	if(mote >= numRemotes || !moteValid[mote])
		throw IllegalWiiMoteIndex();
	return IS_PRESSED(motes[mote], button);
}

bool WiiManager::isButtonPressed(unsigned mote, unsigned button) const {
	if(mote >= numRemotes || !moteValid[mote])
		throw IllegalWiiMoteIndex();
	return IS_JUST_PRESSED(motes[mote], button);
}

bool WiiManager::isButtonReleased(unsigned mote, unsigned button) const{
	if(mote >= numRemotes || !moteValid[mote])
		throw IllegalWiiMoteIndex();
	return IS_RELEASED(motes[mote], button);
}

void WiiManager::connectWiiMotes(){
	wiiuse_find(motes, MAX_REMOTES, 5);
	numRemotes = wiiuse_connect(motes, MAX_REMOTES);

	for (unsigned i = 0; i < min(4u, numRemotes); i++)
		wiiuse_set_leds(motes[i], WIIMOTE_LED_1 * (1 << i));

	for(unsigned i = 0; i < numRemotes; i++) {
		moteValid[i] = true;
		wiiuse_rumble(motes[i], 1);
	}

	usleep(200000);

	for (unsigned i = 0; i < numRemotes; i++)
		wiiuse_rumble(motes[i], 0);
}

void WiiManager::update(){
	if (wiiuse_poll(motes, numRemotes)) {
		for(unsigned i = 0; i < numRemotes; ++i){
			if (motes[i]->event == WIIUSE_UNEXPECTED_DISCONNECT ||
			    motes[i]->event == WIIUSE_DISCONNECT)
				moteValid[i] = false;
		}
	}
}

void WiiManager::attachLimb(unsigned mote, unsigned id, LimbType limb){
	if(mote >= numRemotes || !moteValid[mote])
		throw IllegalWiiMoteIndex();
	limbs[mote]->setLimbAttachment(id, limb);
}

void WiiManager::detachLimb(unsigned mote){
	if(mote >= numRemotes || !moteValid[mote])
		throw IllegalWiiMoteIndex();
	limbs[mote]->removeLimbAttachment();
}
